use core::time::Duration;
use std::collections::HashMap;

use bitfinex::api::Bitfinex;
use bitfinex::positions::Position;
use bitfinex::ticker::TradingPairTicker;
use tokio::time::delay_for;

use crate::connectors::Connector;
use crate::currency::{Symbol, SymbolPair};
use crate::pairs::PairStatus;
use crate::strategy::Strategy;
use crate::ticker::Ticker;
use crate::BoxError;

pub struct BfxBot<'a> {
    connector: Box<dyn Connector + 'a>,
    ticker: Ticker,
    pair_statuses: Vec<PairStatus<'a>>,
    quote: Symbol,
    trading_symbols: Vec<Symbol>,
}

impl<'a> BfxBot<'a> {
    pub fn new<C: Connector + 'a>(
        connector: C,
        trading_symbols: Vec<Symbol>,
        quote: Symbol,
        tick_duration: Duration,
    ) -> Self {
        BfxBot {
            connector: Box::new(connector),
            ticker: Ticker::new(tick_duration),
            pair_statuses: trading_symbols
                .iter()
                .map(|x| SymbolPair::new(quote.clone(), x.clone()))
                .map(|x| PairStatus::new(x, 1, None))
                .collect(),
            quote,
            trading_symbols,
        }
    }

    pub fn with_strategy(mut self, strategy: Box<dyn Strategy>) -> Self {
        self.pair_statuses
            .iter_mut()
            .for_each(|x| x.set_strategy(dyn_clone::clone_box(&*strategy)));

        self
    }

    pub async fn current_prices(&self, symbol: Symbol) -> Result<TradingPairTicker, BoxError> {
        let trading_pair = SymbolPair::new(self.quote.clone(), symbol.clone());

        if !self.trading_symbols.contains(&symbol) {
            return Err("Symbol not supported.".into());
        }

        self.connector.current_prices(&trading_pair).await
    }

    pub async fn start_loop(&mut self) -> Result<(), BoxError> {
        if let Err(e) = self.update_pair_statuses().await {
            println!("Error while updating pairs at first start: {}", e);
        }

        loop {
            self.update().await;
        }
    }

    async fn update(&mut self) {
        delay_for(self.ticker.duration()).await;
        self.ticker.inc();

        if let Err(e) = self.update_pair_statuses().await {
            println!("Error while updating pairs: {}", e);
        }
    }

    async fn update_pair_statuses(&mut self) -> Result<(), BoxError> {
        for status in &mut self.pair_statuses {
            // add positions for each pair
            self.connector
                .active_positions(status.pair())
                .await?
                .into_iter()
                .for_each(|x| status.add_position(x));
        }

        Ok(())
    }
}
