use std::collections::HashMap;

use crate::currency::SymbolPair;
use crate::events::{Event, EventDispatcher, SignalKind};
use crate::orders::Order;
use crate::positions::Position;
use crate::strategy::Strategy;

pub struct PairStatus<'a> {
    pair: SymbolPair,
    dispatcher: EventDispatcher,
    prices: HashMap<u64, f64>,
    events: Vec<Event>,
    orders: HashMap<u64, Vec<Order>>,
    positions: HashMap<u64, Vec<Position>>,
    current_tick: u64,
    strategy: Option<Box<dyn Strategy + 'a>>,
    signals: HashMap<u64, SignalKind>,
}

impl<'a> PairStatus<'a> {
    pub fn new(pair: SymbolPair, current_tick: u64, strategy: Option<Box<dyn Strategy>>) -> Self {
        PairStatus {
            pair,
            dispatcher: EventDispatcher::new(),
            prices: HashMap::new(),
            events: Vec::new(),
            positions: HashMap::new(),
            orders: HashMap::new(),
            signals: HashMap::new(),
            current_tick,
            strategy,
        }
    }

    pub fn dispatcher(&self) -> &EventDispatcher {
        &self.dispatcher
    }
    pub fn dispatcher_mut(&mut self) -> &mut EventDispatcher {
        &mut self.dispatcher
    }

    pub fn add_position(&mut self, position: Position) {
        let (new_position, events, signals) = {
            match &self.strategy {
                Some(strategy) => strategy.position_on_new_tick(&position, &self),
                None => (position, vec![], vec![]),
            }
        };

        self.positions
            .entry(self.current_tick)
            .or_default()
            .push(new_position.clone());

        // calling position state callbacks
        self.dispatcher
            .call_position_state_handlers(&new_position, &self);

        // adding events and calling callbacks
        for e in events {
            self.add_event(e);
        }

        // adding signals to current tick vector
        for s in signals {
            self.add_signal(s);
        }

        println!(
            "EVENTS: {:?} | SIGNALS: {:?} | POSITION: {:?}",
            self.events, self.signals, new_position
        );
    }

    fn add_event(&mut self, event: Event) {
        self.events.push(event);

        self.dispatcher.call_event_handlers(&event, &self);
    }

    fn add_signal(&mut self, signal: SignalKind) {
        self.signals.insert(self.current_tick(), signal);
    }

    pub fn current_tick(&self) -> u64 {
        self.current_tick
    }

    pub fn set_strategy(&mut self, strategy: Box<dyn Strategy + 'a>) {
        self.strategy = Some(strategy);
    }

    pub fn pair(&self) -> &SymbolPair {
        &self.pair
    }

    pub fn position_previous_tick(&self, id: u64, tick: Option<u64>) -> Option<&Position> {
        let tick = match tick {
            Some(tick) => {
                if tick < 1 {
                    1
                } else {
                    tick
                }
            }
            None => self.current_tick() - 1,
        };

        self.positions
            .get(&tick)
            .and_then(|x| x.iter().find(|x| x.position_id() == id))
    }
}
